###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# __init__.py                                                                 #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" Peekaboo constant module data. """


VERSION = (2, 0, 2)

__version__ = '.'.join(map(str, VERSION))
__author__ = 'Felix Bauer'
__description__ = 'Peekaboo Extended Email Attachment Behavior Observation Owl'
__copyright__ = (
    'Copyright (C) 2016-2020 science + computing ag. All rights reserved.')
__license__ = 'GPLv3'


PEEKABOO_OWL = r"""
Peekaboo Extended Email Attachment Behavior Observation Owl

                   _a_aa                    a_aa,
                    '*U4UUUULa_aa_aa_aajUUU4XU7'
                      aX''''''UUXU4XUU'''''!Ua
                    _U'        -U4UU'   _    'U,
                    ?i   jLd1   ?#Wi   4L01   Ui
                    -U,        4#000P        _U'
                     -*Xa_a_a_WUW##KUL_a_a_aX7'
                    _aXUXUUU4UUX4XX444UUUUUUXLa,
                   _UXXUXUXU47'!'!'!'!*X444U4UXX,
                   ?XU4U4''__    ___   ___'UUXUUi
                   ?4U4' |__ \  / _ \ |__ \ 'UUXi
                    *Xi    _) || | | |  _) | ?X7
                     *L   / _/ | |_| | / _/   j7
                      *a |____(_)___(_)____| jY
                       -L,                _/'
                         'l,            _/'
                           j7_a_;  aaa/4
               _aaaaaa#0000#00000##0##00000000aaaaaa,
        aaad0P!!!!!!      Blakiston's Fish       '!!!!!!Laaa
  _aa!!!!                                                    !! _,
(never mind the K)
"""
